#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

#include "parser.h"
#include "io.h"
#include "str.h"


int state = OUT;

/*
 * parse_file(): the main function that parses a netconf
 *               file into a t_netconf
 * 
 * Returns: pointer to t_netconf tree
 * 
 * */

t_netconf *
parse_file (const char *filename) {

	FILE *f;
	char *word, *p, *optname;
	t_netconf *tree = NULL;  /* the tree */
	t_scheme *cs = NULL;     /* current scheme */
	t_iface *ci = NULL;      /*    "    iface */
	t_alias *ca = NULL;      /*    "    alias */
	t_provider *cp = NULL;   /*    "    provider */

	word = malloc (80);
	tree = malloc(sizeof (t_netconf));

	if ((f = fopen (filename, "r")) == NULL) {
		err (strerror (errno));
		return (NULL);
	}

	while (getword (f, word) != NULL) {

		if ((p = strchr (word, '#'))) {  /* take care of comments */
			*p = '\0';
			to_next_line (f);
		}

		if (*word == 0)
			to_next_line (f);

		if (strcmp ("{", word) == 0) {
			switch (state) {
				case READ_DEMAND:
					state = IN_DEMAND;
					break;
				case READ_PROVIDER_NAME:
					state = IN_PROVIDER;
					break;

				case READ_SCHEME_NAME:
					state = IN_SCHEME;
					break;
				case READ_IFACE_NAME:
					state = IN_IFACE;
					break;
				case READ_INET_NAME:
					state = IN_INET;
					break;
			}
		}
		else if (strcmp ("}", word) == 0) {
			switch (state) {
				case READ_INET_OPTION_NAME:
				case READ_INET_OPTION_VAL:
				case IN_INET:
					state = IN_IFACE;
					break;
				case IN_IFACE:
					state = IN_SCHEME;
					break;
				case IN_SCHEME:
					state = OUT;
					break;
				case READ_PROVIDER_OPTION_NAME:
				case READ_PROVIDER_OPTION_VAL:
				case IN_PROVIDER:
					state = IN_DEMAND;
					break;
				case IN_DEMAND:
					state = OUT;
					break;
			}
		}
		else if (state == OUT) {
			if (strcmp ("scheme", word) == 0) {

				state = READ_SCHEME;
				cs = scheme_add (tree->schemes);
				if (tree->schemes == NULL) tree->schemes = cs;
				ci = NULL;
				ca = NULL;
			}	
			else if (strcmp ("demand", word) == 0) {
				state = READ_DEMAND;
			}
		}
		else if (state == READ_SCHEME) {
			state = READ_SCHEME_NAME;
			cs->name = strdup (word);
		}
		else if (state == IN_SCHEME) {
			if (strcmp ("iface", word) == 0) {

				state = READ_IFACE;
				ci = iface_add (cs->iface);
				if (cs->iface == NULL) cs->iface = ci;
				ca = NULL;
			}
		}
		else if (state == READ_IFACE) {
			state = READ_IFACE_NAME;
			ci->name = strdup (word);
		}
		else if (state == IN_IFACE) {
			if (strcmp ("inet", word) == 0) {

				state = READ_INET;
				ca = alias_add (ci->alias);
				if (ci->alias == NULL) ci->alias = ca;
			}
		}
		else if (state == READ_INET) {
			state = READ_INET_NAME;
			setopt (&ca->options, "type", word);
		}
		else if (state == IN_INET) {
			state = READ_INET_OPTION_NAME;
			optname = strdup (word);
		}
		else if (state == READ_INET_NAME) {
			if (strcmp (";", word) == 0) {
				state = IN_IFACE;
			}
		}
		else if (state == READ_INET_OPTION_NAME) {
			setopt (&ca->options, optname, word);
			free (optname); optname = NULL;
			state = READ_INET_OPTION_VAL;
		}
		else if (state == READ_INET_OPTION_VAL) {
			if (strcmp (";", word) == 0) {
				state = IN_INET;
			}
		}
		else if (state == IN_DEMAND) {
			if (strcmp ("provider", word) == 0) {
				state = READ_PROVIDER;
				cp = provider_add (tree->demand);
				if (tree->demand == NULL) tree->demand = cp;
			}
		}
		else if (state == READ_PROVIDER) {
			state = READ_PROVIDER_NAME;
			cp->name = strdup (word);
		}
		else if (state == IN_PROVIDER) {
			state = READ_PROVIDER_OPTION_NAME;
			optname = strdup (word);
		}
		else if (state == READ_PROVIDER_OPTION_NAME) {
			state = READ_PROVIDER_OPTION_VAL;
			setopt (&cp->options, optname, word);
			free (optname); optname = NULL;
		}
		else if (state == READ_PROVIDER_OPTION_VAL) {
			if (strcmp (";", word) == 0) {
				state = IN_PROVIDER;
			}
		}
	}
	return tree;
}


/*
 * scheme_add(): create a new scheme
 * returns: pointer to the original scheme
 *
 * @scheme: pointer to existing t_scheme tree
 *
 * */

t_scheme *
scheme_add (t_scheme *scheme)
{
	t_scheme *ns, *cs = scheme;  /* cs = current scheme; ns = new scheme */

	if (scheme) {
		cs = scheme;
		while (cs->next)
			cs = cs->next;
	}

	ns = malloc (sizeof (t_scheme));

	ns->name = NULL;
	ns->iface = NULL;
	ns->next = NULL;

	if (scheme) {
		cs->next = ns;
	}

	return ns;
}

/*
 * iface_add(): create a new iface
 * returns: pointer to the original t_iface
 *
 * @iface: pointer to existing t_iface tree
 *
 * */

t_iface *
iface_add (t_iface *iface)
{
	t_iface *ni, *ci;    /* ni = new iface; nc = current iface */

	if (iface) {
		ci = iface;
		while (ci->next)
			ci = ci->next;
	}

	ni = malloc (sizeof (t_iface));

	ni->name = NULL;
	ni->alias = NULL;
	ni->next = NULL;

	if (iface) {
		ci->next = ni;
	}

	return ni;
}

/*
 * alias_add(): create a new scheme
 * returns: pointer to the original t_alias
 *
 * @scheme: pointer to existing t_alias tree
 *
 * */

t_alias *
alias_add (t_alias *alias)
{
	t_alias *na, *ca;        /* na = new alias; ca = current alias */

	if (alias) {
		ca = alias;
		while (ca->next)
			ca = ca->next;
	}

	na = malloc (sizeof (t_alias));

	na->next = NULL;
	na->options = NULL;

	if (alias) {
		ca->next = na;
	}

	return na;
}

/*
 * provider_add(): create a new provider
 * returns: pointer to the original t_provider
 *
 * @provider: pointer to existing t_provider tree
 *
 * */

t_provider *
provider_add (t_provider *provider)
{
	t_provider *np, *cp;         /* np = new provider; cp = current provider */

	if (provider) {
		cp = provider;
		while (cp->next)
			cp = cp->next;
	}

	np = malloc (sizeof (t_provider));

	np->next = NULL;
	np->options = NULL;

	if (provider) {
		cp->next = np;
	}

	return np;

}


/*
 * getopt(): get the value of an option inside a t_opt
 *
 * Returns: pointer to option value. It points to the real value inside
 *          the struct t_opt, so don't free() it.
 * 
 * @opt: a pointer to a t_opt containing options
 * @name: the name of the option's value you want to get
 *
 * */

char *
getopt (t_opt *opt, char *name)
{
	t_opt *co = opt;       /* co = current opt; no = new opt */

	if (co == NULL)
		return NULL;

	while (co) {
		if (co->name)
			if (strcmp (co->name, name) == 0)
				return co->val;
		co = co->next;
	}

	return NULL;
}

/*
 * setopt(): set an option inside a t_opt
 * 
 * Returns: nothing
 *
 * @opt: a pointer to a pointer to a t_opt containing the option
 *       you want to set.
 * @name: the option's name
 * @val: the option's new value
 * 
 * */

void
setopt (t_opt **opt, char *name, char *val)
{
	t_opt *co;            /* co = current opt */

	co = *opt;
	while (co) {
		if (co->name) {
			if (strcmp (co->name, name) == 0) {
				free (co->val);
				co->val = strdup (val);
				return;
			}
		}
		if (co->next)
			co = co->next;
		else
			break;
	}

	/* create a new t_opt struct */
	if (*opt == NULL) {
		*opt = malloc (sizeof (t_opt));
		co = *opt;
	}
	else {
		co->next = malloc (sizeof (t_opt));
		co = co->next;
		co->next = NULL;
	}

	/* strdup() the name and val */
	co->name = strdup (name);
	co->val = strdup (val);

	return;
}


void
err (char *str)
{
	errstr = malloc (strlen (str) + 20);
	sprintf (errstr, "error: '%s' at line %d", str, lineno);
}

void
warn (char *str)
{
	

}
